use crate::{fold::impl_fold, kw, Errors, HdlAttr, PairsIterExt};
use proc_macro2::{Span, TokenStream};
use quote::{quote, quote_spanned, ToTokens};
use std::fmt;
use syn::{
    parse::{Parse, ParseStream},
    parse_quote, parse_quote_spanned,
    punctuated::Punctuated,
    token::Paren,
    Attribute, ConstParam, GenericParam, Generics, Ident, ImplGenerics, LifetimeParam, Path,
    PredicateType, Token, Turbofish, Type, TypeGenerics, TypeParam, WhereClause, WherePredicate,
};

crate::options! {
    #[options = TypeOptions]
    pub(crate) enum TypeOption {
        OutlineGenerated(outline_generated),
        ConnectInexact(connect_inexact),
        Target(target, Path),
    }
}

pub(crate) struct WrappedInConst<'a> {
    outer: &'a mut TokenStream,
    span: Span,
    inner: TokenStream,
}

impl<'a> WrappedInConst<'a> {
    pub(crate) fn new(outer: &'a mut TokenStream, span: Span) -> Self {
        Self {
            outer,
            span,
            inner: TokenStream::new(),
        }
    }
    pub(crate) fn inner(&mut self) -> &mut TokenStream {
        &mut self.inner
    }
}

impl Drop for WrappedInConst<'_> {
    fn drop(&mut self) {
        let inner = &self.inner;
        quote_spanned! {self.span=>
            const _: () = {
                #inner
            };
        }
        .to_tokens(self.outer);
    }
}

pub(crate) fn get_target(target: &Option<(kw::target, Paren, Path)>, item_ident: &Ident) -> Path {
    match target {
        Some((_, _, target)) => target.clone(),
        None => item_ident.clone().into(),
    }
}

pub(crate) fn common_derives(span: Span) -> Attribute {
    parse_quote_spanned! {span=>
        #[::fayalite::__std::prelude::v1::derive(
            ::fayalite::__std::fmt::Debug,
            ::fayalite::__std::cmp::Eq,
            ::fayalite::__std::cmp::PartialEq,
            ::fayalite::__std::hash::Hash,
            ::fayalite::__std::marker::Copy,
            ::fayalite::__std::clone::Clone,
        )]
    }
}

crate::options! {
    #[options = LifetimeParamOptions]
    enum LifetimeParamOption {}
}

crate::options! {
    #[options = TypeParamOptions]
    pub(crate) enum TypeParamOption {}
}

crate::options! {
    #[options = ConstParamOptions]
    pub(crate) enum ConstParamOption {}
}

#[derive(Debug, Clone)]
pub(crate) enum ParsedType {}

impl ParsedType {
    pub(crate) fn parse(ty: &mut Type) -> syn::Result<Self> {
        todo!()
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedTypeParam {
    pub(crate) attrs: Vec<Attribute>,
    pub(crate) options: HdlAttr<TypeParamOptions>,
    pub(crate) ident: Ident,
    pub(crate) default: Option<(Token![=], ParsedType)>,
}

#[derive(Debug, Clone)]
pub(crate) enum ParsedGenericParam {
    Type(ParsedTypeParam),
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedGenerics {
    pub(crate) lt_token: Option<Token![<]>,
    pub(crate) params: Punctuated<ParsedGenericParam, Token![,]>,
    pub(crate) gt_token: Option<Token![>]>,
}

impl ParsedGenerics {
    pub(crate) fn parse(generics: &mut Generics) -> syn::Result<Self> {
        let Generics {
            lt_token,
            params,
            gt_token,
            where_clause,
        } = generics;
        let mut errors = Errors::new();
        let mut predicates: Vec<WherePredicate> = Vec::with_capacity(params.len());
        let params = Punctuated::from_iter(params.pairs_mut().filter_map_pair_value_mut(|param| {
            Some(match param {
                GenericParam::Lifetime(param) => {
                    errors.unwrap_or_default(HdlAttr::<LifetimeParamOptions>::parse_and_take_attr(
                        &mut param.attrs,
                    ));
                    errors.error(param, "lifetime generics are not supported by #[hdl]");
                    return None;
                }
                GenericParam::Type(TypeParam {
                    attrs,
                    ident,
                    colon_token,
                    bounds,
                    eq_token,
                    default,
                }) => {
                    let span = ident.span();
                    let options = errors
                        .unwrap_or_default(HdlAttr::<TypeParamOptions>::parse_and_take_attr(attrs))
                        .unwrap_or_default();
                    if !bounds.is_empty() {
                        predicates.push(WherePredicate::Type(PredicateType {
                            lifetimes: None,
                            bounded_ty: parse_quote! { #ident },
                            colon_token: colon_token.unwrap_or_else(|| Token![:](span)),
                            bounds: bounds.clone(),
                        }));
                    }
                    let default = default
                        .as_mut()
                        .and_then(|v| errors.ok(ParsedType::parse(v)));
                    ParsedGenericParam::Type(ParsedTypeParam {
                        attrs: attrs.clone(),
                        options,
                        ident: ident.clone(),
                        default: default.map(|v| (eq_token.unwrap_or_else(|| Token![=](span)), v)),
                    })
                }
                GenericParam::Const(ConstParam {
                    attrs,
                    const_token,
                    ident,
                    colon_token,
                    ty,
                    eq_token,
                    default,
                }) => {
                    let options = errors
                        .unwrap_or_default(HdlAttr::<ConstParamOptions>::parse_and_take_attr(attrs))
                        .unwrap_or_default();
                    todo!()
                }
            })
        }));
        if let Some(where_clause) = where_clause {
            for predicate in &mut where_clause.predicates {
                let WherePredicate::Type(PredicateType {
                    lifetimes,
                    bounded_ty,
                    colon_token,
                    bounds,
                }) = predicate
                else {
                    errors.error(predicate, "unsupported where predicate kind");
                    continue;
                };
                todo!();
            }
        }
        errors.finish()?;
        Ok(Self {
            lt_token: *lt_token,
            params,
            gt_token: *gt_token,
        })
    }
}

impl ToTokens for ParsedGenerics {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        todo!()
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedGenericsImplGenerics<'a>(&'a ParsedGenerics);

impl ToTokens for ParsedGenericsImplGenerics<'_> {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        todo!()
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedGenericsTurbofish<'a>(&'a ParsedGenerics);

impl ToTokens for ParsedGenericsTurbofish<'_> {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        todo!()
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedGenericsTypeGenerics<'a>(&'a ParsedGenerics);

impl ToTokens for ParsedGenericsTypeGenerics<'_> {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        todo!()
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedGenericsWhereClause<'a>(&'a ParsedGenerics);

impl ToTokens for ParsedGenericsWhereClause<'_> {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        todo!()
    }
}

pub trait AsTurbofish {
    type Turbofish<'a>: 'a + ToTokens + Clone + fmt::Debug
    where
        Self: 'a;
    fn as_turbofish(&self) -> Self::Turbofish<'_>;
}

impl AsTurbofish for TypeGenerics<'_> {
    type Turbofish<'a> = Turbofish<'a> where Self: 'a;

    fn as_turbofish(&self) -> Self::Turbofish<'_> {
        TypeGenerics::as_turbofish(self)
    }
}

impl AsTurbofish for ParsedGenericsTypeGenerics<'_> {
    type Turbofish<'a> = ParsedGenericsTurbofish<'a>
    where
        Self: 'a;

    fn as_turbofish(&self) -> Self::Turbofish<'_> {
        ParsedGenericsTurbofish(self.0)
    }
}

pub trait SplitForImpl {
    type ImplGenerics<'a>: 'a + ToTokens + Clone + fmt::Debug
    where
        Self: 'a;
    type TypeGenerics<'a>: 'a + AsTurbofish + ToTokens + Clone + fmt::Debug
    where
        Self: 'a;
    type WhereClause<'a>: 'a + ToTokens + Clone + fmt::Debug
    where
        Self: 'a;
    fn split_for_impl(
        &self,
    ) -> (
        Self::ImplGenerics<'_>,
        Self::TypeGenerics<'_>,
        Self::WhereClause<'_>,
    );
}

impl SplitForImpl for Generics {
    type ImplGenerics<'a> = ImplGenerics<'a>;
    type TypeGenerics<'a> = TypeGenerics<'a>;
    type WhereClause<'a> = Option<&'a WhereClause>;
    fn split_for_impl(
        &self,
    ) -> (
        Self::ImplGenerics<'_>,
        Self::TypeGenerics<'_>,
        Self::WhereClause<'_>,
    ) {
        Generics::split_for_impl(&self)
    }
}

impl SplitForImpl for ParsedGenerics {
    type ImplGenerics<'a> = ParsedGenericsImplGenerics<'a>
    where
        Self: 'a;

    type TypeGenerics<'a> = ParsedGenericsTypeGenerics<'a>
    where
        Self: 'a;

    type WhereClause<'a> = ParsedGenericsWhereClause<'a>
    where
        Self: 'a;

    fn split_for_impl(
        &self,
    ) -> (
        Self::ImplGenerics<'_>,
        Self::TypeGenerics<'_>,
        Self::WhereClause<'_>,
    ) {
        (
            ParsedGenericsImplGenerics(self),
            ParsedGenericsTypeGenerics(self),
            ParsedGenericsWhereClause(self),
        )
    }
}

#[derive(Debug, Clone)]
pub(crate) enum MaybeParsed<P, U> {
    Unrecognized(U),
    Parsed(P),
}

impl<P: ToTokens, U: ToTokens> ToTokens for MaybeParsed<P, U> {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        match self {
            MaybeParsed::Unrecognized(v) => v.to_tokens(tokens),
            MaybeParsed::Parsed(v) => v.to_tokens(tokens),
        }
    }

    fn to_token_stream(&self) -> TokenStream {
        match self {
            MaybeParsed::Unrecognized(v) => v.to_token_stream(),
            MaybeParsed::Parsed(v) => v.to_token_stream(),
        }
    }

    fn into_token_stream(self) -> TokenStream {
        match self {
            MaybeParsed::Unrecognized(v) => v.into_token_stream(),
            MaybeParsed::Parsed(v) => v.into_token_stream(),
        }
    }
}

impl<P: AsTurbofish, U: AsTurbofish> AsTurbofish for MaybeParsed<P, U> {
    type Turbofish<'a> = MaybeParsed<P::Turbofish<'a>, U::Turbofish<'a>>
    where
        Self: 'a;

    fn as_turbofish(&self) -> Self::Turbofish<'_> {
        match self {
            MaybeParsed::Unrecognized(v) => MaybeParsed::Unrecognized(v.as_turbofish()),
            MaybeParsed::Parsed(v) => MaybeParsed::Parsed(v.as_turbofish()),
        }
    }
}

impl<P: SplitForImpl, U: SplitForImpl> SplitForImpl for MaybeParsed<P, U> {
    type ImplGenerics<'a> = MaybeParsed<P::ImplGenerics<'a>, U::ImplGenerics<'a>>
    where
        Self: 'a;
    type TypeGenerics<'a> = MaybeParsed<P::TypeGenerics<'a>, U::TypeGenerics<'a>>
    where
        Self: 'a;
    type WhereClause<'a> = MaybeParsed<P::WhereClause<'a>, U::WhereClause<'a>>
    where
        Self: 'a;

    fn split_for_impl(
        &self,
    ) -> (
        Self::ImplGenerics<'_>,
        Self::TypeGenerics<'_>,
        Self::WhereClause<'_>,
    ) {
        match self {
            MaybeParsed::Unrecognized(v) => {
                let (i, t, w) = v.split_for_impl();
                (
                    MaybeParsed::Unrecognized(i),
                    MaybeParsed::Unrecognized(t),
                    MaybeParsed::Unrecognized(w),
                )
            }
            MaybeParsed::Parsed(v) => {
                let (i, t, w) = v.split_for_impl();
                (
                    MaybeParsed::Parsed(i),
                    MaybeParsed::Parsed(t),
                    MaybeParsed::Parsed(w),
                )
            }
        }
    }
}
